/* A Solution to Experiment 6.3: Build A Web Server Using Sockets
 *
 * David Vos: 1/31/2002
 *
 * Notes:
 *  May not necessarily provide full security (May potentially allow accesses
 *    outside of document root; document root is current path when run;
 *    may contain buffer over-flow issues, etc.)
 *  Formatting of directory listings is not as pretty as it could be.
 *  For sake of simplicity, I cheat and use the 404 error for other kinds of
 *    errors as well.
 ****************************************************************************/
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#include <sys/wait.h>
#include <signal.h>
#include <unistd.h>

#include <time.h>
#include <dirent.h>
#include <sys/stat.h>

#include <stdio.h>
#include <string.h>

#define CGI_PATH "/cgi-bin"

#define P_ERROR(mesg)	{perror(mesg);exit(-1);}

#define	PROTOPORT	4000		/* default protocol port number */
#define	QLEN		6		/* size of request queue	*/

int	visits	    =   0;		/* counts client connections	*/
int	sd, sd2;	 /* socket descriptors			*/
/*------------------------------------------------------------------------
 * Program:   http
 *
 * Syntax:    http [ port ]
 *
 *		 port  - protocol port number to use
 *
 * Note:      The port argument is optional.  If no port is specified,
 *	      the server uses the default given by PROTOPORT.
 *
 *------------------------------------------------------------------------
 */

void get_type(char * ext, char * type);

int fork2();
void handle_req();

/* directory printing functions */
void printdir(char * path);
static int scandir_cp(const char* dir, struct dirent*** namelist,
				int (*mask) (const struct dirent*),
				int (*cmp) (const void *, const void *));
static int alphasort_cp(const void *a, const void *b);
static int mask (const struct dirent* DIRP);

/* Handle ^C */
void sigproc(int i)
{
	struct linger l;

	l.l_onoff = 0;  /* really close connection */
	setsockopt(sd2, SOL_SOCKET, SO_LINGER, (char *) &l, sizeof 1);
	setsockopt(sd, SOL_SOCKET, SO_LINGER, (char *) &l, sizeof 1);
	close(sd2);
	close(sd);
	exit(0);
}

main(argc, argv)
int	argc;
char	*argv[];
{
	struct	hostent	 *ptrh;	 /* pointer to a host table entry	*/
	struct	protoent *ptrp;	 /* pointer to a protocol table entry	*/
	struct	sockaddr_in sad; /* structure to hold server's address	*/
	struct	sockaddr_in cad; /* structure to hold client's address	*/
	int	port;		 /* protocol port number		*/
	int	alen;		 /* length of address			*/
	char	buf[1000];	 /* buffer for string the server sends	*/
	int	reuse = 1;
	pid_t	pid;

	memset((char *)&sad,0,sizeof(sad)); /* clear sockaddr structure	*/
	sad.sin_family = AF_INET;	  /* set family to Internet	*/
	sad.sin_addr.s_addr = INADDR_ANY; /* set the local IP address	*/

	/* Check command-line argument for protocol port and extract	*/
	/* port number if one is specified.  Otherwise, use the	default	*/
	/* port value given by constant PROTOPORT			*/
	if (argc > 1) {			/* if argument specified	*/
		port = atoi(argv[1]);	/* convert argument to binary	*/
	} else {
		port = PROTOPORT;	/* use default port number	*/
	}
	if (port > 0)			/* test for illegal value	*/
		sad.sin_port = htons((u_short)port);
	else {				/* print error message and exit	*/
		fprintf(stderr,"bad port number %s\n",argv[1]);
		exit(1);
	}

	/* Map TCP transport protocol name to protocol number */
	if ( ((int)(ptrp = getprotobyname("tcp"))) == 0)
		P_ERROR("Cannot map tcp to protocol");

	/* Create a socket */
	if ((sd = socket(PF_INET, SOCK_STREAM, ptrp->p_proto)) < 0)
		P_ERROR("Socket creation failed");

	/* Bind a local address to the socket */
	if (setsockopt(sd,SOL_SOCKET, SO_REUSEADDR, (char *) &reuse,
			sizeof(reuse)) < 0 ||
			bind(sd, (struct sockaddr *)&sad, sizeof(sad)) < 0)
		P_ERROR("Bind failed");

	/* Specify size of request queue */
	if (listen(sd, QLEN) < 0)
		P_ERROR("Listen failed");

	signal(SIGINT, sigproc); /* catch ^C */
	/* Main server loop - accept and handle requests */
	while (1) {
		alen = sizeof(cad);
		if ( (sd2=accept(sd, (struct sockaddr *)&cad, &alen)) < 0)
			P_ERROR("Accept failed");
		
		switch (pid = fork2())
		{
		case -1:
			P_ERROR("Fork Failed");
		case 0:
			close(sd);

			handle_req();

			close(sd2);
			_exit(0);
		default:
			close(sd2);
		}
	}
	close(sd);
}


void handle_req()
{
	char	buf[1000], cmd[100], path[100], vers[100], type[30];
	FILE	*file_ptr;
	struct stat statbuffer;
	int i;

	/* Read request */
	memset(buf,0,1000);
	recv(sd2,buf,1000,0);

	/* Decifer request */
	sscanf(buf,"%99s %99s %99s",cmd,path,vers);

	/* Handle a request for '/' */
	if (strcmp(path, "/") == 0)
		sprintf(path, "/.");

	do_cgi(path);    /* Don't return if cgi : ) */

	/* check if file exists */
	if (stat(path+1, &statbuffer) == -1)
		error_404(path+1);

	/* if it is a directory print its contents, otherwise send the file */
	if (S_ISDIR(statbuffer.st_mode))
		printdir(path + 1);
	else {
		file_ptr=fopen(path+1,"r");

		if (file_ptr) {
			get_type(strrchr(path, '.'), type);
			sprintf(buf,"HTTP/1.1 200 Document follows\r\nServer: cpscProj\r\nContent-type: %s\r\n\r\n", type);
			send(sd2,buf,strlen(buf),0);

			while ((i=fread(buf,1,1000,file_ptr))>0) {
				send(sd2,buf,i,0);
			}
		}
		else
			error_404(path);

		fclose(file_ptr);
	}
}


/* Get the content-type for a given extension */
void get_type(char * ext, char * type)
{
	if (ext == NULL)
		goto def;

	ext++; /*check after period */

	switch (*ext){
	case 'h':
		if (strncmp(ext, "htm", 3) == 0)
			sprintf(type, "text/html");
		else goto def;
		break;
	case 't':
		if (strncmp(ext, "tif", 3) == 0)
			sprintf(type, "image/tiff");
		else if (strncmp(ext, "txt", 3) == 0)
			sprintf(type, "text/plain");
		else goto def;
		break;
	case 's':
		if (strncmp(ext, "sgml", 4) == 0)
			sprintf(type, "text/sgml");
		else goto def;
		break;
	case 'm':
		if (strncmp(ext, "mpg", 3) == 0 || strncmp(ext, "mpeg", 4) == 0)
			sprintf(type, "video/mpeg");
		else goto def;
		break;
	case 'j':
		if (strncmp(ext, "jpg", 3) == 0 || strncmp(ext, "jpeg", 4) == 0)
			sprintf(type, "image/jpeg");
		else goto def;
		break;
	case 'p':
		if (strncmp(ext, "png", 3) == 0)
			sprintf(type, "image/png");
		else if (strncmp(ext, "ps", 2) == 0)
			sprintf(type, "application/postscript");
		else goto def;
		break;
	case 'b':
		if (strncmp(ext, "bmp", 3) == 0)
			sprintf(type, "image/x-bitmap");
		else goto def;
		break;
	case 'g':
		if (strncmp(ext, "gif", 3) == 0)
			sprintf(type, "image/gif");
		else goto def;
		break;
	case '\0':
		sprintf(type, "text/plain");
		break;
	default:
	def:
		/* The default type under RFC 1945 */
		sprintf(type, "application/octet-stream");
	}
}


/* fork2() -- like fork, but the new process is immediately orphaned
 *            and won't leave a zombie when it exits. */
int fork2()
{
	pid_t pid;
	int rc;
	int status;

	if (!(pid = fork()))
	{
		switch (fork())
		{
			case 0:	return 0;
			case -1: _exit(-1);
			default: _exit(0);
		}
	}

	if (pid < 0 || waitpid(pid, &status, 0) < 0)
		return -1;

	if (WIFEXITED(status) && WEXITSTATUS(status) == 0)
			return 1;

	return -1;
}


void printdir(char * path)
{
	int n, i;
	struct dirent** dirlist;
	char buf[1000];
	char ppath[100];
	struct stat statbuffer, statfile;
	char mtime[30];
	int topdir = 0;
	char * slash;
	char filename[100];

	if ((n=scandir_cp(path, &dirlist, &mask, alphasort_cp)) < 0) {
		fprintf(stderr, "Error reading directory.\n");
		exit(-1);
	}

	sprintf(ppath, "/%s", path);
	if (strlen(ppath) <= 1 || strcmp(ppath, "/.") == 0)
		topdir = 1;
	else
		if (ppath[strlen(ppath)-1] == '/')  /* remove trailing slash */
			ppath[strlen(ppath)-1] = '\0';

	slash = strrchr(ppath, '/');
	if (slash != NULL)
		slash[1] = '\0'; /* cut off rest of string */
		
	/* send header stuff */
	sprintf(buf,"HTTP/1.1 200 Document follows\r\nServer: cpscProj\r\nContent-type: text/html\r\n\r\n <HTML><HEAD><TITLE>Index of /%s</TITLE></HEAD>\r\n<BODY><H1>Index of /%s</H1>\r\n<pre><IMG SRC=\"/blank.gif\"> Name\t\tLast Modified\n\t\t\t\t\tSize<hr>\r\n", path, path);
	send(sd2,buf,strlen(buf),0);

	if (!topdir) {
		sprintf(buf,"<img src=\"/back.gif\"> <a href=\"\%s\">Parent Directory</a>\r\n", ppath);
		send(sd2,buf,strlen(buf),0);
	}

	for (i = 0; i < n; i++) {
		sprintf(filename, "%s/%s", path, dirlist[i]->d_name);
		if(stat(filename, &statbuffer) == -1)
			continue;

		if (S_ISDIR(statbuffer.st_mode))
			sprintf(buf, "<IMG SRC=\"/folder.gif\" ALT=\"[   ]\"> <A HREF=\"%s/\">%s/</A>\t%s\t\t\t\t\t%d bytes\r\n", dirlist[i]->d_name, dirlist[i]->d_name, ctime(&(statbuffer.st_mtime)), statbuffer.st_size);
		else
			sprintf(buf, "<IMG SRC=\"/unknown.gif\" ALT=\"[   ]\"> <A HREF=\"%s\">%s</A>\t%s\t\t\t\t\t%d bytes\r\n", dirlist[i]->d_name, dirlist[i]->d_name, ctime(&(statbuffer.st_mtime)), statbuffer.st_size);

		send(sd2,buf,strlen(buf),0);
	}
	sprintf(buf,"</body></html>\r\n\r\n");
	send(sd2,buf,strlen(buf),0);
}


/* return 0 if not a cgi request, otherwise it shouldn't return at all. : )
   if it is a cgi request, do it. */
int do_cgi(char * path)
{
	char *cgi_path = CGI_PATH;
	char *cmd_path;
	char ch_path[100];
	char *cmd_exe;
	char *cmd_args;
	char cmd_var1[100];
	char cmd_var2[100];

	if (strncmp(path, cgi_path, sizeof(cgi_path) - 1) != 0)
		return 0;

	cmd_path = path + 1; /* don't include leading '/' */

	cmd_exe = strrchr(path, '/');  /* find begining of command */
	cmd_args = strchr(path, '?');	/* find arguments after command */

	if (cmd_exe == NULL)
		error_404("");

	if (cmd_exe == NULL || strlen(cmd_exe) <= 1)
		error_404("Bad Command name");

	/* delimit parts of command */
	*(cmd_exe) = '\0';
	cmd_exe++;	/* jump to beginning of command */
	if (cmd_args != NULL) {
		*(cmd_args) = '\0';
	}

	/* change to cgi script path */
	sprintf(ch_path, "./%s", cmd_path);
	if (chdir(ch_path) < 0)
		error_404("path");

	/* Setup CGI environment variables */
	/* Note: can be easily modified to pass in SERVER_NAME or
			REMOTE_ADDR...*/
	if (cmd_args != NULL)
	{
		cmd_args++;	/* jump to beginning of arguments */
		sprintf(cmd_var1, "QUERY_STRING=%s", cmd_args); /* must remain in memory */
		if (putenv(cmd_var1) != 0) {
			perror("putenv() failed");
			_exit(-1);
		}
	}
	sprintf(cmd_var2, "REQUEST_METHOD=GET"); /* must remain in memory */
	if (putenv(cmd_var2) != 0) {
		perror("putenv() failed.");
		_exit(-1);
	}

	/* run the command (and don't come back) */
	exec_cgi(cmd_exe, sd2);

	error_404(cmd_exe);
}


/* Execute cgi script like the exec() command, except handle output of script */
/* Function does not return */
int exec_cgi(char * cmd, int socket_descripter)
{
	FILE * fpipe;
	char buf[1000], header[1000];
	int i;

	/* Run command and pipe output into fpipe */
	sprintf(buf, "./%s", cmd);
	if ((fpipe = popen(buf, "r")) == NULL) {
		perror("Unable to execute cgi program");
		_exit(-1);
	}

	if (fgets(buf, 1000, fpipe) == NULL)
		error_404("Error executing cgi program");

	if (strncasecmp(buf, "Location:", 9) == 0) {
		/* return 302 FOUND message */
		buf[strlen(buf) - 1] = '\0'; /* remove \n from string (replace with \r\n) */

		/* double-check for buffer-overflow */
		if (strlen(buf) >= 200)
			error_404("Malformed 302 FOUND message");

		/* Message taken from Apache output */
		sprintf(header, "HTTP/1.1 302 Found\r\nServer: cpscProj\r\n%s\r\nContent-Type: text/html; charset=iso-8859-1\r\n\r\n<!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML 2.0//EN\">\r\n<HTML><HEAD>\r\n<TITLE>302 Found</TITLE>\r\n</HEAD><BODY>\r\n<H1>Found</H1>\r\nThe document has moved <A HREF=\"%s\"here</A>.<P>\r\n</BODY></HTML>\r\n\r\n", buf, buf+10);

		send(socket_descripter,header,strlen(header),0);
	}
	/* Note: could handle other situations here, also */
	else {
		sprintf(header, "HTTP/1.1 200 OK\r\nServer: cpscProj\r\n");
		send(socket_descripter,header,strlen(header),0);
		send(sd2,buf,strlen(buf),0);

		// send rest of output
                while (fgets(buf, 1000, fpipe) != NULL) {
                        send(sd2,buf,strlen(buf),0);
                }
	}

	if (pclose(fpipe) < 0) {
                        perror("pclose() failed");
                        _exit(-1);
	}
	_exit(-1);
}


/* Send an HTTP 404 error */
/* Function does not return */
int error_404(char * path)
{
        char buf[1000];

        sprintf(buf,"HTTP/1.1 404 File Not Found\r\nServer: cpscProj\r\nContent-type: text/html\r\n\r\n<html><head><title>Not Found</title></head><body><h1>404 Not Found</h1>The requested URL %s was not found on this server.</body></html>\r\n\r\n", path);
        send(sd2,buf,strlen(buf),0);

        _exit(-1);
}


/* For some reason <sys/dir.h> is missing on the computer holmes, so Paul Maki
   put this together for our cpsc 247 project.  --David */
/* Due to an issue with alphasort on holmes, I have to make an almost
   identical copy.  (Read: I am "borrowing" this code). */
static int scandir_cp(const char* dir, struct dirent*** namelist,
		int (*mask) (const struct dirent*),
		int (*cmp) (const void *, const void *))
{
	DIR *dp = opendir(dir);
	struct dirent **v = NULL;
	size_t vsize = 0, i;
	struct dirent *d;
	int save;

	if (dp == NULL)
		return -1;

	i = 0;
	while((d = readdir(dp)) != NULL)
		if(mask == NULL || (*mask)(d)) {
			struct dirent *vnew;
			size_t dsize;

			if (i == vsize) {
				struct dirent **new;
				if (vsize == 0)
					vsize = 10;
				else
					vsize *= 2;
				new = (struct dirent **)realloc(v, vsize*sizeof(*v));
				if (new == NULL)
					break;
				v = new;
			}

			dsize = &d->d_name[strlen((d)->d_name)+1] - (char *) d;
			vnew = (struct dirent *) malloc(dsize);
			if (vnew == NULL)
				break;

			v[i++] = (struct dirent *)memcpy(vnew, d, dsize);
		}

	(void) closedir(dp);

	if (cmp != NULL)
		qsort(v, i, sizeof (*v), cmp);
	*namelist = v;
	return i;
}


/* sorts a list alphabetically */
static int alphasort_cp(const void *a, const void *b)
{
	return strcoll((*(const struct dirent **) a)->d_name,
		       (*(const struct dirent **) b)->d_name);
}

/* used to exclude files from a directory during a call to scandir() */
static int mask(const struct dirent* DIRP)
{
	/* ignore hidden files */
	if((strncmp(".", DIRP->d_name, 1))==0)
		return 0;
	return 1;
}
